@@ -168,17 +168,18 @@ def pset_hash(
     # [1] https://death.andgravity.com/stable-hashing
     # [2] https://ourpython.com/python/deterministic-recursive-hashing-in-python
     # [3] https://stackoverflow.com/a/52175075
-    try:
-        if skip_special_cols:
-            keys = [x for x in dct.keys() if not x.startswith("_")]
-            _dct = {kk: dct[kk] for kk in keys}
-        else:
-            _dct = dct
-        return joblib.hash(_dct, hash_name=method)
+    if skip_special_cols:
+        _dct = {
+            key: val for key, val in dct.items() if not key.startswith("_")
+        }
+    else:
+        _dct = dct
     # joblib can hash "anything" so we didn't come up with an input that
     # actually fails to hash. As such, TypeError is just a guess here. But
     # still we don't catch ValueError raised when an invalid hash_name is
     # passed (anything other than md5 or sha1).
+    try:
+        return joblib.hash(_dct, hash_name=method)
     except TypeError as ex:
         if raise_error:
             raise PsweepHashError(
@@ -188,6 +189,17 @@ def pset_hash(
             return np.nan
 
 
+def add_hashes(params: Sequence[dict]) -> Sequence[dict]:
+    """Calculate hash for each pset and set pset["_pset_hash"]. Return new
+    params list.
+    """
+    new_params = []
+    for pset in params:
+        pset["_pset_hash"] = pset_hash(pset)
+        new_params.append(pset)
+    return new_params
+
+
 def check_calc_dir(calc_dir: str, df: pd.DataFrame):
     """Check calc dir for consistency with database.
 
@@ -660,27 +672,55 @@ def pgrid(plists):
     return itr2params(itertools.product(*plists))
 
 
-def filter_same_hash(params: Sequence[dict], **kwds) -> Sequence[dict]:
+def filter_params_unique(params: Sequence[dict], **kwds) -> Sequence[dict]:
     """Reduce params to unique psets.
 
+    Use pset["_pset_hash"]  if present, else calculate hash on the fly.
+
     Parameters
     ----------
-    params :
-    kwds :
+    params
+    kwds
         passed to :func:`pset_hash`
     """
-    msk = np.unique(
-        [pset_hash(dct, **kwds) for dct in params], return_index=True
-    )[1]
+    get_hash = lambda pset: pset.get("_pset_hash", pset_hash(pset, **kwds))
+    msk = np.unique([get_hash(pset) for pset in params], return_index=True)[1]
     return [params[ii] for ii in np.sort(msk)]
 
 
+def filter_same_hash(*args, **kwds):
+    warnings.warn(
+        "filter_same_hash() was renamed to filter_params_unique()",
+        DeprecationWarning,
+    )
+    return filter_params_unique(*args, **kwds)
+
+
+def filter_params_dup_hash(
+    params: Sequence[dict], hashes: Sequence[str], **kwds
+) -> Sequence[dict]:
+    """Return params with psets whose hash is not in `hashes`.
+
+    Use pset["_pset_hash"]  if present, else calculate hash on the fly.
+
+    Parameters
+    ----------
+    params
+    hashes
+    kwds
+        passed to :func:`pset_hash`
+    """
+    get_hash = lambda pset: pset.get("_pset_hash", pset_hash(pset, **kwds))
+    return [pset for pset in params if not get_hash(pset) in hashes]
+
+
 def stargrid(
     const: dict,
     vary: Sequence[dict],
     vary_labels: Sequence[str] = None,
     vary_label_col: str = "_vary",
-    filter_dups=True,
+    filter_dups=None,
+    skip_dups=True,
 ) -> Sequence[dict]:
     """
     Helper to create a specific param sampling pattern.
@@ -689,10 +729,10 @@ def stargrid(
     values (middle of the "star").
 
     When doing that, duplicate psets can occur. By default try to filter them
-    out (use filter_same_hash()) but ignore hash calculation errors and return
+    out (use filter_params_unique()) but ignore hash calculation errors and return
     non-reduced params in that case. If you want to fail at hash errors, use
 
-    >>> filter_same_hash(stargrid(..., filter_dups=False), raise_error=True)
+    >>> filter_params_unique(stargrid(..., skip_dups=False), raise_error=True)
 
     Examples
     --------
@@ -710,7 +750,7 @@ def stargrid(
      {'a': 1, 'b': 88, 'c': 11},
      {'a': 1, 'b': 99, 'c': 11}]
 
-    >>> ps.stargrid(const, vary=[a, b], filter_dups=False)
+    >>> ps.stargrid(const, vary=[a, b], skip_dups=False)
     [{'a': 1, 'b': 77, 'c': 11},
      {'a': 2, 'b': 77, 'c': 11},
      {'a': 3, 'b': 77, 'c': 11},
@@ -749,9 +789,14 @@ def stargrid(
                 _dct = dct
             params.append(merge_dicts(const, _dct))
 
-    if filter_dups:
+    if filter_dups is not None:
+        skip_dups = filter_dups
+        warnings.warn(
+            "filter_dups was renamed to skip_dups", DeprecationWarning
+        )
+    if skip_dups:
         try:
-            return filter_same_hash(
+            return filter_params_unique(
                 params, raise_error=True, skip_special_cols=True
             )
         except PsweepHashError:
@@ -782,6 +827,10 @@ def worker_wrapper(
     pset_seq=np.nan,
     run_seq: int = None,
 ):
+    """
+    Add special fields to pset. Call worker on exactly one pset. Return
+    DataFrame row built from ``pset.update(worker(pset))``.
+    """
     assert run_id is not None
     assert calc_dir is not None
     pset_id = str(uuid.uuid4())
@@ -792,7 +841,6 @@ def worker_wrapper(
         "_pset_id": pset_id,
         "_calc_dir": calc_dir,
         "_time_utc": time_start,
-        "_pset_hash": pset_hash(pset, PSET_HASH_ALG, raise_error=False),
         "_pset_seq": pset_seq,
         "_run_seq": run_seq,
     }
@@ -828,8 +876,12 @@ def run_local(
     database_basename="database.pk",
     backup=False,
     git=False,
+    skip_dups=False,
 ) -> pd.DataFrame:
     """
+    Call `worker` for each `pset` in `params`. Populate a DataFrame with rows
+    from each call ``worker(pset)``.
+
     Parameters
     ----------
     worker : Callable
@@ -845,7 +897,9 @@ def run_local(
         * None : use serial execution
         * int : use multiprocessing.Pool (even for ``poolsize=1``)
     save : bool
-        save final DataFrame to ``<calc_dir>/database.pk`` (pickle format only)
+        save final DataFrame to ``<database_dir>/<database_basename>`` (pickle
+        format only), default: "calc/database.pk", see also `database_dir`,
+        `calc_dir` and `database_basename`
     tmpsave : bool
         save results from each `pset` in `params` (the current DataFrame row) to
         ``<calc_dir>/tmpsave/<run_id>/<pset_id>.pk`` (pickle format only)
@@ -854,6 +908,9 @@ def run_local(
         * sequence : list of DataFrame column names, print the row but only
           those columns
     calc_dir : str
+        Dir where calculation artifacts can be saved if needed, such as dirs
+        per pset ``<calc_dir>/<pset>``. Will be added to the database in
+        ``_calc_dir`` field.
     simulate : bool
         run everything in ``<calc_dir>.simulate``, don't call `worker`, i.e. save
         what the run would create, but without the results from `worker`,
@@ -861,13 +918,16 @@ def run_local(
     database_dir : str
         Path for the database. Default is ``<calc_dir>``.
     database_basename : str
-        ``<database_dir>/<database_basename>``
+        ``<database_dir>/<database_basename>``, default: "database.pk"
     backup : bool
         Make backup of ``<calc_dir>`` to ``<calc_dir>.bak_<timestamp>_run_id_<_run_id>``
     git : bool
         Use ``git`` to commit all files written and changed by the current run
         (``_run_id``). Make sure to create a ``.gitignore`` manually before if
         needed.
+    skip_dups : bool
+        Skip psets whose hash is already present in `df`. Useful when repeating
+        (parts of) a study.
     """
 
     database_dir = calc_dir if database_dir is None else database_dir
@@ -916,6 +976,11 @@ def run_local(
         )
         shutil.copytree(calc_dir, dst)
 
+    params = add_hashes(params)
+
+    if skip_dups and len(df) > 0:
+        params = filter_params_dup_hash(params, df._pset_hash.values)
+
     run_id = str(uuid.uuid4())
 
     worker_wrapper_partial = partial(
@@ -1037,6 +1102,7 @@ def prep_batch(
     git: bool = False,
     backup: bool = False,
     write_pset: bool = False,
+    skip_dups: bool = False,
 ) -> pd.DataFrame:
 
     """
@@ -1058,7 +1124,14 @@ def prep_batch(
                 pickle_write(pj(calc_dir, pset["_pset_id"], "pset.pk"), pset)
         return {}
 
-    df = run_local(worker, params, calc_dir=calc_dir, backup=backup, git=False)
+    df = run_local(
+        worker,
+        params,
+        calc_dir=calc_dir,
+        backup=backup,
+        skip_dups=skip_dups,
+        git=False,
+    )
 
     msk_latest = df._run_seq == df._run_seq.values.max()
     msk_old = df._run_seq < df._run_seq.values.max()
